/*
 * socketDevice.c : unix socket device control for use with abpoll,
 * 		    a new active badge station network poller.
 */

#include <errno.h>
#include <stdio.h>
#ifdef _DOS
#include <sys\types.h>
#include <netdb.h>
#include <netinet\in.h>
#include <sys\socket.h>
#include <netinet\in.h>
#include <sys\time.h>
#include <sys\ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <sys\timeb.h>
#include <telnet\protrtn.h>
#include <io.h>
#else
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <termio.h>
#endif

#include "socket.h"
#include "abmalloc.h"

#define SOCKET_READ_MESSAGE_SIZE	(768)   /* size of buffer passed to read() */
#define QUEUE_LENGTH			(1)	/* 1 outstanding socket connection */
#define PAUSE				(5)	/* seconds between socket operation attempts */

#define ACCEPT_TIMEOUT	((long) 20) 	/* usecs timeout on select for accept */
#define READ_TIMEOUT	((long) 20) 	/* usecs timeout on select for read */
#define WRITE_TIMEOUT	((long) 3) 	/* secs timeout on select for write */

#define NULL_LENGTH	(0)
#define NULL_NETWORK	(255)
#define	NULL_TYPE	(0)
#define NULL_STATION	(0)

#ifndef TRUE
#define TRUE            1
#endif

#ifndef FALSE
#define FALSE           0
#endif

typedef struct _SocketList {
    int                 	socket;
    unsigned int		counter;
    struct _SocketMessage	*message;
    struct _SocketList  	*next;
} SocketList;

typedef struct _PortList {
    int				port_number;
    int				listen_socket;
    unsigned long      		number_of_connections;
    unsigned long		tcp_sent;
    struct _SocketList  	*sockets;
    struct _PortList		*next; 
    struct _SocketMessage	*messages;
    struct _SocketMessage     	*lastMessage;
} PortList;

static PortList         *ports = (PortList *) NULL;

static int		socket_debug = FALSE;

static unsigned char	socketReadMessage[SOCKET_READ_MESSAGE_SIZE];

/*
 * declaration of static functions 
 */

#ifdef __STDC__

#ifdef _DOS
static void		sleep(unsigned long);
#endif
static void		addMessage(PortList *, SocketList *);
static int		socketAddConnection(PortList *, int);
static int		socketDeleteConnection(PortList *, int);
static int		socketDeletePort(int);
static int		socketStoreBytesRX(PortList *, SocketList *, unsigned int, unsigned char *);
static int		socketWrite(int, unsigned char *, int);
static int		socketRead(PortList *, SocketList *); 
static int		socketMaintain(PortList *);

#else /* __STDC__ */

#ifdef _DOS
static void		sleep();
#endif
static void		addMessage();
static int		socketAddConnection();
static int		socketDeleteConnection();
static int		socketDeletePort();
static int		socketStoreBytesRX();
static int		socketWrite();
static int		socketRead();
static int		socketMaintain();

#endif /* __STDC__ */


/*
 * static functions 
 */

#ifdef _DOS
static void
sleep(interval)
    unsigned long	interval;
{
    struct timeb	now, timer;

    ftime(&timer);
    do {
        ftime(&now);
    }while ((unsigned long) ((now.time-timer.time)*1000 + now.millitm-timer.millitm) < interval);
}
#endif

static void
addMessage(pPtr, sPtr)
    PortList		*pPtr;
    SocketList		*sPtr;
{
    if (pPtr == (PortList *) NULL) {
	fprintf(stderr, "socket.addMessage: error - can't add message to null port\n");
	return;
    }

    if (sPtr->message == (SocketMessage *) NULL) {
    	fprintf(stderr, "socket.addMessage: error - can't add null message\n");
    	return;
    }

    sPtr->message->next = (SocketMessage *) NULL;

    if (socket_debug) {
    	int		i;
    	
    	fprintf(stderr, "socket.addMessage: message from %d.%d, to net.stn %d.%d, length %d    \n", pPtr->port_number, sPtr->message->socket, sPtr->message->network, sPtr->message->station, sPtr->message->length);
    	
    	for (i = 0; i < (int) (sPtr->message->length & 0xff); i++) {
    	    fprintf(stderr, "%02x ", sPtr->message->body[i]);
    	}
    	fprintf(stderr, "    \n");
    }
    
    if (pPtr->messages == (SocketMessage *) NULL) {
	pPtr->messages = sPtr->message;
	pPtr->lastMessage = sPtr->message;
    }
    else {
	pPtr->lastMessage->next = sPtr->message;
	pPtr->lastMessage = sPtr->message;
    }
}
	

static int
socketAddConnection(pPtr, socket)
    PortList	*pPtr;
    int		socket;
{
    SocketList		*ptr, *new;

    for (ptr = pPtr->sockets; ptr != (SocketList *) NULL; ptr = ptr->next) {
	if (ptr->socket == socket) {
	    break;
	}
    }

    if (ptr != (SocketList *) NULL) {
	fprintf( stderr, "socket:addSocket %d : already exists\n", socket);
	return FALSE;
    }

    MALLOC(new,SocketList,1,"socket.socketAddConnection");

    new->socket = socket;
    new->counter = 0;
    new->message = (SocketMessage *) NULL;

    new->next = pPtr->sockets;
    pPtr->sockets = new;

    pPtr->number_of_connections++;
    
    return TRUE;
}

static int
socketDeleteConnection(pPtr, socket)
    PortList	*pPtr;
    int		socket;
{
    SocketList		*ptr, **list;

    ptr = (SocketList *) NULL;
    list = &(pPtr->sockets);

    while (*list) {
	if ((*list)->socket == socket) {
	    ptr = *list;
	    
	    if (socket_debug) printf("deleting socket %d\n", socket);
	    
	    *list = (*list)->next;
	    break;
	}
	else {
	    list = &((*list)->next);
	}
    }

    if (ptr == (SocketList *) NULL) {
    	
	fprintf(stderr, "socket:socketDeleteConnect(%d) : unknown socket \n", socket);
	
	return FALSE;
    }

    if (shutdown(ptr->socket, 2) != 0) {
    	
    	fprintf(stderr, "socket: warning: failed to shutdown socket %d\n", ptr->socket);

    }
    
    if (close(ptr->socket) != 0) {
    	
        fprintf(stderr, "socket: warning: failed to close socket %d\n", ptr->socket);
        
    }


    if (ptr->message != (SocketMessage *) NULL) {
	FREE(ptr->message);
    }
    FREE(ptr);
 
    pPtr->number_of_connections--;

    return TRUE;
}

static int
socketDeletePort(port_number)
    int		port_number;
{
    PortList	**list, *pPtr;
    
    if (socket_debug) fprintf(stderr, "socketDeletePort for port %d\n", port_number); 
  
    list = &(ports);
    pPtr = (PortList *) NULL;
     
    while (*list) {
 	if ((*list)->port_number == port_number) {
 	    pPtr = (*list);
 	    (*list) = (*list)->next;
 	    break;
 	}
 	else {
 	    list = &((*list)->next);
 	}
    }
    
    if (pPtr == (PortList *) NULL) {
    	fprintf(stderr, "socketDeletePort for unnknown port %d      \n",port_number);
    	return FALSE;
    }
    
    /* we close all socket connections on this port */
       	
    while (pPtr->sockets != (SocketList *) NULL) {
    	if (socket_debug) fprintf(stderr, "socketDeletePort: deleting socket %d on port %d\n", pPtr->sockets->socket, pPtr->port_number);
    	socketDeleteConnection(pPtr, pPtr->sockets->socket);
    }

    /* close the listen socket */   
    
    if (shutdown(pPtr->listen_socket, 2) != 0) {
    	    	
    	fprintf(stderr, "socket: warning: failed to shutdown listen socket %d\n", pPtr->listen_socket);

    }
    
    if (close(pPtr->listen_socket) != 0) {
    	
        fprintf(stderr, "socket: warning: failed to close listen socket %d\n", pPtr->listen_socket);
        
    }
 
    FREE(pPtr); 
    
    return TRUE;
      
}

static int
socketStoreBytesRX(pPtr, sPtr, mLength, mBody)
    PortList		*pPtr;
    SocketList		*sPtr;
    unsigned int	mLength;
    unsigned char	*mBody;
{
    unsigned int	index;
    unsigned int	copy_length;    
    static int		fsm_debug = FALSE;

    if (socket_debug) {
	int		i;

	fprintf(stderr,"socketStoreBytesRX : %d bytes for %d.%d    \n", mLength, pPtr->port_number, sPtr->socket);
	for (i = 0; i < (unsigned int) mLength; i++) {
	    fprintf(stderr, "%02x", mBody[i]);
	}
	fprintf(stderr, "    \n");
    }
      
    for (index = 0; index < mLength; index++) {
    	
        if (sPtr->message == (SocketMessage *) NULL) { /* no previous msg to complete */

            MALLOC(sPtr->message,SocketMessage,1,"socketStoreBytesRX.message");

            sPtr->message->socket = sPtr->socket;
                        
            sPtr->message->length = NULL_LENGTH;
            
            sPtr->message->network = NULL_NETWORK;

            sPtr->message->body[0] = NULL_TYPE;

            sPtr->counter = 0;				/* counter of how much body we've got so far */
            
            sPtr->message->station = mBody[index];	/* first byte in msg is stn */
            
	    if (mBody[index] == NULL_STATION) {
	    	if (socket_debug) fprintf(stderr, "error reading packet from port %d socket %d - bad station\n", pPtr->port_number, sPtr->socket);
            	return FALSE;
            }

	    if (fsm_debug) fprintf(stderr, "new message from %d.%d - station %d    \n", pPtr->port_number, sPtr->socket, sPtr->message->station);
        }

        else if (sPtr->message->body[0] == NULL_TYPE) { /* next byte is the type byte */

            sPtr->message->body[0] = mBody[index];
            
            if (mBody[index] == NULL_TYPE) {
	    	if (socket_debug) fprintf(stderr, "error reading packet from port %d socket %d - bad type byte\n", pPtr->port_number, sPtr->socket);
            	return FALSE;
            }

	    if (fsm_debug) fprintf(stderr, "new message from %d.%d - type %02x    \n", pPtr->port_number, sPtr->socket, sPtr->message->body[0]);
        }
        else if (sPtr->message->length == NULL_LENGTH) { /* next byte is the length byte */
          
            sPtr->message->length = (unsigned int) (mBody[index] +  2); /* +2 because _this_ length count includes the type byte and the length byte */

            sPtr->message->body[1] = mBody[index];

            if (mBody[index] == NULL_LENGTH) {
           	if (socket_debug) fprintf(stderr, "error reading packet from port %d socket %d - zero length byte\n", pPtr->port_number, sPtr->socket);
		return FALSE;
	    }

	    if (socket_debug) fprintf(stderr, "new message from %d.%d - length %02x    \n", pPtr->port_number, sPtr->socket, sPtr->message->body[1]);
	}
	else if (sPtr->counter < (unsigned int) (sPtr->message->body[1] & 0xff)) { /* if we have not filled the full body of the message yet */
	   
	    if ((((unsigned int) (sPtr->message->body[1] & 0xff)) - sPtr->counter) <= mLength - index) { /* there are enough bytes left to finish the body of this message */
	        copy_length = (unsigned int) (sPtr->message->body[1] & 0xff) - sPtr->counter;
	    }
	    else { /* not enough bytes left (yet) to fill the body of this message */
	
	        copy_length = mLength - index;
	    }

	    memcpy( &(sPtr->message->body[2 + sPtr->counter]), &(mBody[index]), copy_length);

	    index += (copy_length - 1); /* `-1' because at the end of this for loop, index will get incremented by one */
	    sPtr->counter += copy_length;

	    if (fsm_debug) fprintf(stderr, "message from %d.%d - copyed %d bytes, awaiting %d more (counter is %d), and then the network number    \n", pPtr->port_number, sPtr->socket, copy_length, sPtr->message->body[1] - sPtr->counter, sPtr->counter);

	}
	else if (sPtr->message->network == NULL_NETWORK) {
		
	    sPtr->message->network = mBody[index];

	    if (fsm_debug) fprintf(stderr, "new message from %d.%d - network %d    \n", pPtr->port_number, sPtr->socket, sPtr->message->network);
	    
	    if (mBody[index] == NULL_NETWORK) {    
           	if (socket_debug) fprintf(stderr, "error reading packet from port %d socket %d - zero length byte\n", pPtr->port_number, sPtr->socket);
		return FALSE;
	    }
	    
	    addMessage(pPtr, sPtr);
	    sPtr->message = (SocketMessage *) NULL; 
	}
	else {
	    if (socket_debug) fprintf(stderr, "FSM in socketProcessBytesRX got screwed - socket %d, station %d, network %d, length %d\n", sPtr->message->socket, sPtr->message->station, sPtr->message->network, sPtr->message->length);
	    return FALSE;
	}
    }

    return TRUE;
}	       



	
static int
socketWrite(length, message, socket)
    int			length;
    unsigned char	*message;
    int			socket;
{
    int			nfds, writefds, result, n, counter;
    struct timeval	timeout;
    unsigned char	*message_ptr;

    message_ptr = message;
    counter = length;

    for (;;) {
    	
        timeout.tv_sec = (long) WRITE_TIMEOUT;;
        timeout.tv_usec = (long) 0;
        
        nfds = socket + 1;
        writefds = 1 << socket;
        
        result = select( nfds, (fd_set *) 0, (fd_set *) &writefds, (fd_set *) 0, &timeout);
        
        if (result > 0) { /* select succeeded */
        
            if ((writefds & (1 << socket)) != 0) { /* socket available for writing */
            
                n = write(socket, message_ptr, counter);

                if (n == counter) {
                    return TRUE;
                }
                else if (n <= 0) {
	            fprintf(stderr, "socketWrite : socket %d returned %d\n", socket, n);
	            perror("");
	            return(FALSE);
        	} else {
            	    counter -= n;

        	    if (socket_debug) fprintf(stderr, "partial write to socket %d: %d remaining\n", socket, counter);

            	    message_ptr = &(message_ptr[n]);
		}
            }
            else { /* select lied! (should never happen) */
                fprintf(stderr, "socket: select for writing lied about socket %d\n", socket);
                return FALSE;
            }
        }
	else if (result < 0) { /* select failed */
	    fprintf(stderr, "socket: select for writing failed for socket %d\n", socket);
	    return FALSE;
	}
	else { /* select timed out */
	    fprintf(stderr, "socket: select for writing timed out on socket %d\n", socket);
	    return FALSE;
	}
    }
}

/*
 *	socketRead returns < 0 when the socket that was read from dies,
 *	otherwise it returns the number of bytes read from the socket.
 */

static int
socketRead(pPtr, sPtr)
    PortList	*pPtr;
    SocketList	*sPtr;
{

    int            		nfds, readfds;
    struct timeval      	timeout;
    int				length, result;

    timeout.tv_sec = (long) 0;
    timeout.tv_usec = (long) READ_TIMEOUT;

    nfds = sPtr->socket + 1;
    readfds = 1 << sPtr->socket;

    result = select(nfds, (fd_set*) &readfds, (fd_set *) 0, (fd_set *) 0, &timeout);

    if (result > 0) { /* select succeeded */

	if ((readfds & (1 << sPtr->socket)) != 0) { /* something to read */
	
 	    if ((length = read(sPtr->socket, socketReadMessage, SOCKET_READ_MESSAGE_SIZE)) <= 0) {
		if (socket_debug) {
	            fprintf(stderr, "socket: socket read failed\n");
	            perror("");
	        }	    
		return -1;
	    }
	}
	else { /* select lied! (should never happen) */
	    fprintf(stderr, "socket: select lied about socket %d\n", sPtr->socket);
	    return -1;
	}

    }
    else if (result < 0) { /* select failed */
	fprintf(stderr, "socket: select failed for socket %d\n", sPtr->socket);
	return -1;
    }
    else { /* select timed out */
	return 0; /* nothing to read */
    }
   
    /** so, we've more than zero bytes from the socket - good stuff **/

    if (socketStoreBytesRX(pPtr, sPtr, length, socketReadMessage) == FALSE) {
	
	if (socket_debug) {
	    int		i;

	    fprintf(stderr, "socket: socketStoreBytesRX for %d.%d failed for %d bytes \n", pPtr->port_number, sPtr->socket, length);
	    
	    fprintf(stderr, "<< ");
	    for (i = 0; i < length; i++) {
		fprintf(stderr, "%02x ", socketReadMessage[i]);
	    }
	    fprintf(stderr, ">> \n");
	}
	
	/** if the protocol between the poller and the slave get's 
	    screwed, then there is no re-synchronization, so the best
	    that we can do is to close the connection with the slave, and
	    let it re-open it and hope that it talks properly next time 
	 **/
	
	return -1;
    }

    return length; /** we return the number of bytes read **/
}    


/*
 * Return FALSE if the listen socket has gone bad and should be closed
 * Return TRUE if everything is still ok
 */

static int
socketMaintain(pPtr)
    PortList	*pPtr;
{
    int				new_socket;
    struct timeval      	timeout;
    int                 	optval = 1;
    int                 	optlen = sizeof(optval);
    static struct sockaddr_in   addr;
    int				addr_len = sizeof(struct sockaddr_in);
    int				n = -1, nfds;
    long			readfds;
    int				result;
    
    nfds = pPtr->listen_socket + 1;
    
    readfds = 1 << pPtr->listen_socket;
    
    timeout.tv_sec = 0;
    timeout.tv_usec = ACCEPT_TIMEOUT;

    result = select(nfds, (fd_set*) &readfds, (fd_set *) 0, (fd_set *) 0, &timeout);

    if (result > 0) {    	

	if (socket_debug) printf("socketMaintain: select succeeded\n");
	
	if (readfds & (1 << pPtr->listen_socket)) {

	    if (socket_debug) fprintf(stderr, "socketMaintain: about to accept on listen socket %d.%d     \n", pPtr->port_number, pPtr->listen_socket);

	    if ((new_socket = accept(pPtr->listen_socket, (struct sockaddr *)&addr, &addr_len)) < 0) {
		fprintf(stderr, "listen socket accept failed on port %d errno %d\n", pPtr->port_number, errno);
		return FALSE;
	    }

	    setsockopt(new_socket, SOL_SOCKET, SO_REUSEADDR, &optval, optlen);

	    socketAddConnection(pPtr, new_socket);
	    return TRUE;
	}
	else { /* select lied! (should never happen) */
	    fprintf(stderr, "socket: select lied about socket %d\n", pPtr->listen_socket);
	    return FALSE;
	}
    }
    else if (result < 0) {
        if (socket_debug) printf("socketMaintain: select failed\n");
	return FALSE;
    }

/**    	
    if (socket_debug) printf("socketMaintain: select timed out\n");
**/    	
    return TRUE;
}

/* -------------------------------------------------------------------
 * public functions
 */

int
socketNumberOfConnections(port_number, number)
    int		port_number;
    int		*number;
{
    PortList	*pPtr;

    for (pPtr = ports; pPtr != (PortList *) NULL; pPtr = pPtr->next) {
    	if (pPtr->port_number == port_number) {
    	    break;
    	}
    }
   
    if (pPtr == (PortList *) NULL) {
    	return FALSE;
    }
    
    *number = (int) pPtr->number_of_connections;
    
    return TRUE;
}

int
socketInitialise(port_number)
    int		port_number;
{
    PortList		*ptr;
    struct sockaddr_in	sin;
    int                 arg = 1;
    int                 	optval = 1;
    int                 	optlen = sizeof(optval);

    for (ptr = ports; ptr != (PortList *) NULL; ptr = ptr->next) {
    	if (ptr->port_number == port_number) {
    	    break;
    	}
    }
    
    if (ptr != (PortList *) NULL) {
    	fprintf(stderr, "Port %d already initialised, skipping ...\n", port_number);
    	return TRUE;
    }
    
    MALLOC(ptr,PortList,1,"socket.socketInitalise"); 
      
    ptr->port_number = port_number;
    ptr->sockets     = (SocketList *) NULL;
    ptr->messages    = (SocketMessage *) NULL;
    ptr->lastMessage = (SocketMessage *) NULL;
    ptr->number_of_connections = 0;
    ptr->tcp_sent    = 0; /* compatibility with built-in display version */
    ptr->next        = ports;
    ports            = ptr;

    if ((ptr->listen_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	fprintf(stderr, "could not acquire listen socket\n");
	perror("");
	exit(1);
    }

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port_number);


    /*
     * mark the listen socket as non-blocking
     */

#ifdef _dos
    
    if (ioctl(ptr->listen_socket, FIONBIO, &arg) < 0) {
    	fprintf(stderr, "ioctl FIONBIO on listen socket on port %d failed\n", port_number);
	exit(1);
    }
#else
    {
	int	n;
	
	n = fcntl(ptr->listen_socket, F_SETFL, O_NDELAY);
	if (n < 0) {
	    fprintf(stderr, "failed to set non-blocking io on port %d\n", port_number);
	    exit(1);
	}
    }
#endif


    if (setsockopt(ptr->listen_socket, SOL_SOCKET, SO_REUSEADDR, &optval, optlen) != 0) {
	fprintf(stderr, "setsockopt failed\n");
	exit(1);
    }

    while (bind(ptr->listen_socket, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
	fprintf(stderr, "listen socket on port %d: bind failed\n", port_number);
	perror("");
	sleep(PAUSE);
    }

    while (listen(ptr->listen_socket, QUEUE_LENGTH) < 0) {
	fprintf(stderr, "socket listen failed on port %d\n", port_number);
	perror("");
	sleep(PAUSE);
    }

    return TRUE;
}

int
socketTX(port_number, socket_number, length, message)
    int			port_number;
    int			socket_number; /* socket number -1 if TXsion to occur to all sockets on this port */
    int			length;
    unsigned char	*message;
{
    PortList	*pPtr;
    SocketList	*sPtr;
    int		dead_socket;

    for (pPtr = ports; pPtr != (PortList *) NULL; pPtr = pPtr->next) {
        if (pPtr->port_number == port_number) {
            break;
        }	
    }

    if (pPtr == (PortList *) NULL) {
    	fprintf(stderr, "socketTX to unknown port %d\n", port_number);
    	return FALSE;
    }

    if (socket_number != -1) {
    	for (sPtr = pPtr->sockets; sPtr != (SocketList *) NULL; sPtr = sPtr->next) {
    	    if (sPtr->socket == socket_number) {
    	    	break;
    	    }
    	}
    	
        if (sPtr == (SocketList *) NULL) {
            return FALSE;
        }
	
	if (!(socketWrite(length, message, sPtr->socket))) {
		
	    if (socket_debug) fprintf(stderr, "socketDevice: write failed to socket %d\n", sPtr->socket);
	    
	    dead_socket = sPtr->socket;
	    sPtr = sPtr->next;
	    socketDeleteConnection(pPtr, dead_socket);
	}
	else {
	    pPtr->tcp_sent++;
	    sPtr = sPtr->next;
	}

	return TRUE;
    }

    sPtr = pPtr->sockets;
    while (sPtr != (SocketList *) NULL) {
    	if (socket_debug) printf("about to socketWrite %d bytes to %d.%d      \n", length, pPtr->port_number, sPtr->socket); 
	if (!(socketWrite(length, message, sPtr->socket))) {
	    if (socket_debug) {		
	        fprintf(stderr, "socketDevice: write failed to socket %d\n", sPtr->socket);
	    }
	    dead_socket = sPtr->socket;
	    sPtr = sPtr->next;
	    socketDeleteConnection(pPtr, dead_socket);
	}
	else {
	    pPtr->tcp_sent++;
	    sPtr = sPtr->next;
	}
    }

    return TRUE;
}
 
SocketMessage *
socketRX(port_number)
    int			port_number;
{
    PortList		*pPtr;
    SocketList		*sPtr;
    SocketMessage	*mPtr;
    int			result;

    for (pPtr = ports; pPtr != (PortList *) NULL; pPtr = pPtr->next) {
    	if (pPtr->port_number == port_number) {
    	    break;
    	}
    }
    
    if (pPtr == (PortList *) NULL) {
        fprintf(stderr, "socketRX on unknown port %d\n", port_number);
        return (SocketMessage *) NULL;	
    }

    /* look out for new incoming socket connections to this port */

    if (socketMaintain(pPtr) == FALSE) { /* the listen socket has gone bad */

       if (socket_debug) fprintf(stderr, "socketMaintain failed for port %d\n", pPtr->port_number);
       
       socketDeletePort(pPtr->port_number);
	       
       if (!(socketInitialise(port_number))) {
       	  fprintf(stderr, "socketMaintain failed, failed to reinitialize listen socket on port %d\n", port_number);
       }
       
       return (SocketMessage *) NULL;
    }

    /* return a message if one has been received */

    if (pPtr->messages != (SocketMessage *) NULL) {
	
	mPtr = pPtr->messages;
	
	pPtr->messages = pPtr->messages->next;
	
	return mPtr; /** the caller of this function _must_ free this memory **/
    }

    /* poll every socket connection for an incoming message */

    sPtr = pPtr->sockets;
    while (sPtr != (SocketList *) NULL) {

	result = socketRead(pPtr, sPtr);

	if (result < 0) {
	    int 	dead_socket;
	    
	    if (socket_debug) fprintf(stderr, "socketDevice: read failed from socket %d\n", sPtr->socket);
	    
	    dead_socket = sPtr->socket;
	    sPtr = sPtr->next;

	    socketDeleteConnection(pPtr, dead_socket);
	}
	else {
	    sPtr = sPtr->next;
	}
    }

    /* return a message if one has been received */

    if (pPtr->messages != (SocketMessage *) NULL) {

	mPtr = pPtr->messages;
	
	pPtr->messages = pPtr->messages->next;

	return mPtr;  /** the caller of this function _must_ free this memory **/
    }

    return (SocketMessage *) NULL;
}    

unsigned long
socketTCPSent(port_number)
    int    port_number;
{
    PortList	*pPtr;
    
    for (pPtr = ports; pPtr != (PortList *) NULL; pPtr = pPtr->next) {
        if (pPtr->port_number == port_number) {
            break;
        }	
    }
    
    if (pPtr == (PortList *) NULL) {
    	return 0;
    }
    
    return (pPtr->tcp_sent);
}
